from dataclasses import dataclass, field
from typing import List, Dict, Tuple

import pandas as pd

from minotor.data_managers.data_containers.single_feature_data_container import SingleFeatureDataContainer
from minotor.data_managers.data_containers.values_informations import ValuesInformations
from minotor.data_managers.data_types import DataType


@dataclass
class FeaturesDataContainer:
    features: Dict[str, SingleFeatureDataContainer] = field(default_factory=dict)
    values_infos: ValuesInformations = field(default_factory=ValuesInformations)

    @staticmethod
    def from_json(data: Dict):
        return FeaturesDataContainer(
            features={
                feature_name: SingleFeatureDataContainer.from_json(feature)
                for feature_name, feature in data["features"].items()
            },
            values_infos=ValuesInformations.from_json(data["valuesInfos"])
        )

    def get_dict(self):
        return {"features": {feature_name: feature.get_dict() for feature_name, feature in self.features.items()},
                "valuesInfos": self.values_infos.get_dict()
                }

    def update_feature(self, feature_name, data: Dict, data_type: DataType, is_training: bool):
        if feature_name not in self.features:
            self._add_feature(feature_name, data_type)
        self.features[feature_name].update_feature(data, is_training)

    def add_values_infos(self, values_ids: List, values_dates: List, is_training: bool):
        self.values_infos.add_ids(values_ids, is_training)
        self.values_infos.add_dates(values_dates, is_training)

    def _add_feature(self, feature_name: str, data_type: DataType):
        self.features[feature_name] = SingleFeatureDataContainer(data_type)

    def get_dataframes(self, feature_names: List[str] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
        if not feature_names:
            feature_names = self.features.keys()
        return pd.DataFrame({name: self.features[name].training_phase.list_values for name in feature_names},
                            index=self.values_infos.training.ids), \
               pd.DataFrame({name: self.features[name].prediction_phase.list_values for name in feature_names},
                            index=self.values_infos.prediction.ids)



import functools
from typing import Callable, Union

import numpy as np
import pandas as pd

from minotor.data_managers.file_manager import FileManager
from minotor.data_managers.preprocessors import type2preprocessor


def monitor_training_features(func: Callable):
    return _decorator_monitor_features_factory(func, training=True)


def monitor_prediction_features(func: Callable):
    return _decorator_monitor_features_factory(func, training=False)


def _decorator_monitor_features_factory(func: Callable, training: bool = False):
    functools.wraps(func)

    def wrapper_data(data: Union[np.ndarray, pd.DataFrame], *args, **kwargs):
        allowed_types_string = ' '.join(
            [str(elt) for elt in type2preprocessor.keys()])
        assert type2preprocessor.get(type(data)) is not None, \
            f"Your data type must be one of the following : {allowed_types_string}"
        file_manager = FileManager()
        feature_data_container = file_manager.get_features_data()
        preprocessor = type2preprocessor[type(data)]
        for feature_name, feature_data, data_type in preprocessor.preprocess(data):
            feature_data_container.update_feature(
                feature_name, feature_data, data_type, training)
        feature_data_container.add_values_infos(
            preprocessor.current_ids, preprocessor.current_dates, training)
        file_manager.write_features_data(feature_data_container)
        return func(data, *args, **kwargs)

    return wrapper_data



import json

from minotor.encoders.json_encoder_library import ENCODERS


class ExtendedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        for instance, func in ENCODERS:
            if isinstance(obj, instance):
                return func(obj)
        return super(ExtendedJSONEncoder, self).default(obj)



import functools
from typing import Callable

import requests

from minotor.data_managers.file_manager import FileManager
import time


def monitor_predictions(func: Callable):
    functools.wraps(func)

    def wrapper_data(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        duration = time.perf_counter() - start_time
        duration_per_example = duration / len(result)
        file_manager = FileManager()
        prediction_data_container = file_manager.get_prediction_data()
        prediction_data_container.update_predictions(
            result, duration_per_example)
        file_manager.write_prediction_data(prediction_data_container)
        return result

    return wrapper_data



from enum import Enum


class DataType(Enum):
    INT = "int"
    FLOAT = "float"
    CATEGORY = "category"
    BOOL = "bool"
    OTHER = "other"

    @staticmethod
    def type2value(dtype) -> 'DataType':
        for element in DataType:
            try:
                if dtype == element.value:
                    return element
            except TypeError:
                continue
        return DataType.OTHER



from typing import List, Union, Tuple

import numpy as np
import pandas as pd

from pandas.core.indexes.base import Index
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE


class Projector:
    def __init__(self, projector: Union[TransformerMixin, BaseEstimator]):
        self.projector = projector

    @staticmethod
    def preprocess(df: pd.DataFrame) -> pd.DataFrame:
        return df.select_dtypes('number').dropna()

    def project(self, training_df: pd.DataFrame, prediction_df: pd.DataFrame) -> Tuple[Index, Index, np.ndarray, np.ndarray]:
        """
        :return: kept_training_ids, kept_prediction_ids, projected training vectors, projected prediction vectores
        """
        training_df = self.preprocess(training_df)
        prediction_df = self.preprocess(prediction_df)
        full_df = pd.concat([training_df, prediction_df])
        if len(full_df) < 2:
            return [], [], [], []
        projection = self.projector.fit_transform(full_df.to_numpy())
        return training_df.index, prediction_df.index, \
            projection[: len(training_df)], projection[len(training_df):]


tsne_projector = Projector(TSNE())
pca_projector = Projector(PCA(n_components=2))



import json

from tornado.escape import json_decode

from minotor.api.base_route import BaseRouteHandler
from minotor.encoders.json_encoder import ExtendedJSONEncoder
from minotor.data_managers.file_manager import FileManager
from minotor.statistics.projection import tsne_projector


class ProjectionHandler(BaseRouteHandler):
    def post(self):
        if self.request.headers['Content-Type'] == 'application/json':
            feature_names = json_decode(self.request.body)
            fm = FileManager()
            cached_data = fm.get_features_data()
            training_df, prediction_df = cached_data.get_dataframes(
                feature_names)
            training_ids, prediction_ids, training_projection, prediction_projection = tsne_projector.project(
                training_df, prediction_df)
            self.set_header('Content-Type', 'application/json')
            self.write(json.dumps(
                {
                    "training": {"values": training_projection, "ids": training_ids},
                    "prediction": {"values": prediction_projection, "ids": prediction_ids}
                }, cls=ExtendedJSONEncoder))
            self.set_status(200)
        else:
            self.set_status(
                400, 'Request should have Content-Type set to application/json')



from minotor.api.base_route import BaseRouteHandler
from minotor.decorators.feature_monitoring import monitor_training_features
import pandas as pd
import pickle


@monitor_training_features
def save_data(data):
    pass


class TrainingDataHandler(BaseRouteHandler):
    def post(self):
        try:
            data = pickle.loads(self.request.files["training_data"][0]["body"])
        except Exception as e:
            self.set_status(400, "You must send a pickled file")
            return
        try:
            save_data(data)
        except Exception as e:
            self.set_status(400, str(e))
        return



from dataclasses import dataclass, field
from typing import List, Dict

from minotor.statistics.statistic_library import statistic_library, StatisticLibrary


@dataclass
class PredictionData:
    data: Dict = field(default_factory=lambda: {
        "predictions": {
            "values": [],
            "timing": []
        }
    })

    def update_predictions(self, data: List, duration_per_example: float):
        self._add_prediction_values(data, [duration_per_example] * len(data))
        self._compute_predictions_statistics(statistic_library)

    def _add_prediction_values(self, data: List, timing: List):
        self.data["predictions"]["values"].extend(data)
        self.data["predictions"]["timing"].extend(timing)

    def _compute_predictions_statistics(self, statistics: StatisticLibrary):
        for statistic_name, result in statistics.compute_all_statistics(self.data["predictions"]["values"]):
            self.data["predictions"][statistic_name] = result



from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Dict

from minotor.constants import DATETIME_FORMAT


@dataclass
class ValuesInformationsPerPhase:
    ids: List[str] = field(default_factory=list)
    dates: List[datetime] = field(default_factory=list)

    @staticmethod
    def from_json(data: Dict):
        return ValuesInformationsPerPhase(ids=data["ids"],
                                          dates=[datetime.strptime(date_string, DATETIME_FORMAT) for date_string in
                                                 data["dates"]])

    def get_dict(self):
        return {
            "ids": self.ids,
            "dates": [date.strftime(DATETIME_FORMAT) for date in self.dates]
        }


@dataclass
class ValuesInformations:
    training: ValuesInformationsPerPhase = field(default_factory=ValuesInformationsPerPhase)
    prediction: ValuesInformationsPerPhase = field(default_factory=ValuesInformationsPerPhase)

    @staticmethod
    def from_json(data: Dict):
        return ValuesInformations(training=ValuesInformationsPerPhase.from_json(data["training"]),
                                  prediction=ValuesInformationsPerPhase.from_json(data["prediction"]))

    def get_dict(self):
        return {
            "training": self.training.get_dict(),
            "prediction": self.prediction.get_dict(),
            "id2phase": {
                **{id: "training" for id in self.training.ids},
                **{id: "prediction" for id in self.prediction.ids}
            }
        }
    
    def add_ids(self, values_ids: List, is_training: bool):
        if is_training:
            self.training.ids.extend(values_ids)
        else:
            self.prediction.ids.extend(values_ids)

    def add_dates(self, values_dates: List, is_training: bool):
        if is_training:
            self.training.dates.extend(values_dates)
        else:
            self.prediction.dates.extend(values_dates)



import json
import os
from pathlib import Path
from typing import Dict

from minotor.constants import DATA_DIR
from minotor.data_managers.data_containers.features_data_container import FeaturesDataContainer
from minotor.data_managers.prediction_data import PredictionData
from minotor.encoders.json_encoder import ExtendedJSONEncoder


class FileManager:
    def __init__(self):
        self.feature_json_path: Path = DATA_DIR / "feature_data.json"
        self.prediction_json_path: Path = DATA_DIR / "prediction_data.json"

    def get_features_data(self) -> FeaturesDataContainer:
        return FeaturesDataContainer.from_json(
            _load_json(self.feature_json_path)) if self.feature_json_path.exists() else FeaturesDataContainer()

    def get_json(self) -> Dict:
        return _load_json(self.feature_json_path) \
            if self.feature_json_path.exists() else FeaturesDataContainer().get_dict()

    def write_features_data(self, project_data: FeaturesDataContainer):
        with self.feature_json_path.open('w') as f:
            json.dump(project_data.get_dict(), f, cls=ExtendedJSONEncoder)

    def get_prediction_data(self) -> PredictionData:
        return PredictionData(
            _load_json(self.prediction_json_path)
        ) if self.prediction_json_path.exists() else PredictionData()

    def write_prediction_data(self, project_data: PredictionData):
        with self.prediction_json_path.open('w') as f:
            json.dump(project_data.data, f, cls=ExtendedJSONEncoder)

    def clean_data(self):
        self.clean_feature_data()
        self.clean_prediction_data()

    def clean_feature_data(self):
        if os.path.exists(self.feature_json_path):
            os.remove(self.feature_json_path)

    def clean_prediction_data(self):
        if os.path.exists(self.prediction_json_path):
            os.remove(self.prediction_json_path)


def _load_json(path: Path):
    with path.open('r') as f:
        return json.load(f)



from tornado import web

class BaseRouteHandler(web.RequestHandler):
  def options(self):
    # no body
    self.set_status(204)
    self.finish()

  def set_default_headers(self):
    self.set_header("Access-Control-Allow-Origin", "*")
    self.set_header("Access-Control-Allow-Headers", "*")
    self.set_header('Access-Control-Allow-Methods', " *")



from abc import abstractmethod, ABC
from dataclasses import dataclass, field
from typing import Dict, List

from minotor.statistics.statistic_library import StatisticLibrary


@dataclass
class FeaturePhaseContainerABC(ABC):
    values: Dict = field(default_factory=dict)
    statistics: Dict = field(default_factory=dict)

    def get_dict(self):
        return dict({"values": self.values}, **self.statistics)

    @property
    def list_values(self) -> List:
        return list(self.values.values())

    @abstractmethod
    def add_values(self, data: Dict):
        pass

    def compute_statistics(self, statistic_library: StatisticLibrary):
        for statistic_name, result in statistic_library.compute_all_statistics(self.list_values):
            self.statistics[statistic_name] = result


@dataclass
class FeatureTrainingPhaseContainer(FeaturePhaseContainerABC):
    def add_values(self, data: Dict):
        self.values = data

    @staticmethod
    def from_json(data: Dict):
        return FeatureTrainingPhaseContainer(values=data["values"],
                                             statistics={stat_name: val
                                                         for stat_name, val in data.items() if stat_name != "values"})


@dataclass
class FeaturePredictionPhaseContainer(FeaturePhaseContainerABC):
    def add_values(self, data: Dict):
        self.values = {**data, **self.values}

    @staticmethod
    def from_json(data: Dict):
        return FeaturePredictionPhaseContainer(values=data["values"],
                                               statistics={stat_name: val
                                                           for stat_name, val in data.items() if stat_name != "values"})



import webbrowser

from tornado import ioloop, web

from minotor.api.projection_handler import ProjectionHandler
from minotor.api.data_handler import DataHandler
from minotor.api.training_data_handler import TrainingDataHandler

from minotor.constants import PACKAGE_PATH

# Defining constants
REACT_BUILD_PATH = PACKAGE_PATH / 'front' / 'build'
STATIC_PATH = REACT_BUILD_PATH / 'static'


class DefaultHandler(web.RequestHandler):
    def get(self):
        self.render(str(REACT_BUILD_PATH / "index.html"))


def make_app():
    return web.Application([
        (r"/data", DataHandler),
        (r"/projection", ProjectionHandler),
        (r"/training-data", TrainingDataHandler),
        (r"/static/(.*)", web.StaticFileHandler,
         {'path': STATIC_PATH, 'default_filename': 'index.html'}),
        (r"/*.*", DefaultHandler),
    ], debug=True)


def runserver():
    url = 'http://localhost:8888'
    print('Starting server ...')
    app = make_app()
    app.listen(8888)
    print(f'Minotor is available at {url} in your browser !')
    webbrowser.open_new_tab(url)
    ioloop.IOLoop.current().start()


if __name__ == "__main__":
    runserver()



from typing import List, Tuple, Callable, Any, Dict, Generator

import numpy as np
from scipy.stats import entropy


class StatisticLibrary:
    def __init__(self):
        self.statistics: Dict[str, Callable[[np.ndarray], Any]] = {}

    def add_statistic(self, statistic: Callable[[np.ndarray], Any], name: str = None):
        if name is None:
            name = statistic.__name__
        self.statistics[name] = statistic

    def compute_all_statistics(self, values: List[List]) -> Generator[Any, None, None]:
        for statistic_name in self.statistics:
            yield statistic_name, self._compute_statistic(values, statistic_name)

    def _compute_statistic(self, values: List, statistic_name: str):
        statistic = self.statistics[statistic_name]
        try:
            arr = np.array(values).astype("float")
        except ValueError:
            return None
        return statistic(arr)


def _histogram(values: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    histogram_w_arrays = np.histogram(values[~np.isnan(values)], bins="auto")
    return histogram_w_arrays[0], histogram_w_arrays[1]


def _nan_percentage(values: np.ndarray) -> float:
    return sum(np.isnan(values)) / len(values)


def percentile_factory(percentile: int):
    return lambda x: np.nanpercentile(x, percentile)


def _compute_distribution(data: np.ndarray, bins: np.ndarray, epsilon=0.00001) -> np.ndarray:
    # adding epsilon enables us to handle null probabilities
    return np.histogram(data[~np.isnan(data)], bins=bins)[0] + epsilon


def _compute_bins(data: np.ndarray) -> np.ndarray:
    return np.histogram(data[~np.isnan(data)], bins="auto")[1]


def compute_kl_divergence(data: np.ndarray):
    """
    :param data: an array containing the training data as first element and production data as second element
    """
    training_data = data[0]
    prod_data = data[1]
    bins = _compute_bins(np.concatenate([training_data, prod_data]))
    p = _compute_distribution(training_data, bins)
    q = _compute_distribution(prod_data, bins)
    return entropy(p, q)


# Regular Statistic Library

statistic_functions = {
    "mean": np.nanmean,
    "std": np.nanstd,
    "hist": _histogram,
    "nanPercentage": _nan_percentage,
    "percentile95": percentile_factory(95),
    "percentile05": percentile_factory(5)
}
statistic_library = StatisticLibrary()
for name, func in statistic_functions.items():
    statistic_library.add_statistic(func, name)

# Statistic Library using predict and train
crossed_statistic_library = StatisticLibrary()
crossed_statistic_library.add_statistic(compute_kl_divergence, "KLDivergence")



import datetime
from abc import abstractmethod, ABC
from typing import List, Tuple, Dict

import numpy as np
import pandas as pd

from minotor.constants import DATETIME_ID_FORMAT
from minotor.data_managers.data_types import DataType


class PreprocessorABC(ABC):
    def __init__(self):
        self.current_ids = None
        self.current_dates = None

    def preprocess(self, data) -> List[Tuple[str, Dict, str]]:
        """
        :param data: data under the input format
        :return: A list corresponding to the features under the format : [(feature_name, values, data_type)]
        """
        self.current_ids, self.current_dates = self.get_values_infos(data)
        return self.get_preprocessed_data(data)

    @abstractmethod
    def get_values_infos(self, data) -> Tuple[List, List]:
        """
        :param data: data under the input format
        :return: values ids and values dates
        """
        pass

    @abstractmethod
    def get_preprocessed_data(self, data) -> List[Tuple[str, Dict, str]]:
        pass


class NumpyArrayPreprocessor(PreprocessorABC):
    def get_values_infos(self, data: np.ndarray) -> Tuple[List, List]:
        current_date = datetime.datetime.now()
        return [f"{current_date.strftime(DATETIME_ID_FORMAT)}-{i}" for i in range(data.shape[0])], \
               [current_date for _ in range(data.shape[0])]

    def get_preprocessed_data(self, data: np.ndarray) -> List[Tuple[str, Dict, DataType]]:
        data = data.transpose()
        return [(f"feature_{i}",
                 {key: val for key, val in zip(self.current_ids, _replace_nan_with_none(feature_data))},
                 DataType.type2value(feature_data.dtype))
                for i, feature_data in enumerate(data)]


class PandasDataFramePreprocessor(PreprocessorABC):
    def get_values_infos(self, data: pd.DataFrame) -> Tuple[List, List]:
        current_date = datetime.datetime.now()
        return [f"{current_date.strftime(DATETIME_ID_FORMAT)}-{id}" for id in data.index], \
               [current_date for _ in range(len(data))]

    def get_preprocessed_data(self, data: pd.DataFrame) -> List[Tuple[str, Dict, DataType]]:
        return [(col,
                {key: val for key, val in zip(self.current_ids, _replace_nan_with_none(data[col].values))},
                 DataType.type2value(data[col].dtype))
                for col in data]


type2preprocessor = {
    np.ndarray: NumpyArrayPreprocessor(),
    pd.DataFrame: PandasDataFramePreprocessor()
}


def _replace_nan_with_none(array: np.ndarray) -> List:
    return [x if not pd.isna(x) else None for x in array]



import itertools

import numpy as np
import pandas as pd

NUMPY_ENCODINGS = [
    (np.integer, int),
    (np.floating, float),
    (np.ndarray, lambda obj: obj.tolist()),
]

PANDAS_ENCODINGS = [
    (pd.core.indexes.base.Index, lambda obj: obj.tolist())
]

ENCODERS = NUMPY_ENCODINGS + PANDAS_ENCODINGS



import setuptools
setuptools.setup(
    name="minotor",
    version="0.0.2",
    author="Datarmada",
    author_email="contact@datarmada.com",
    description="Easy to use plug and play ML monitoring tool.",
    url="https://github.com/datarmada/minotor",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        "requests>=2.23.0",
        "numpy>=1.18.1",
        "pandas>=1.0.1",
        "tornado>=6.0.4",
        "scikit-learn>=0.22.2",
        "scipy>=1.4.1",
        "wheel>=0.34.2",
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'minotor=minotor.command:main',
        ],
    },
)



import numpy as np
import pandas as pd
from random import uniform

from minotor.data_managers.file_manager import FileManager
from minotor.decorators.feature_monitoring import monitor_training_features, monitor_prediction_features

fm = FileManager()


@monitor_training_features
def train(data):
    return data


@monitor_prediction_features
def predict(data):
    return data


def run():
    fm.clean_feature_data()

    # Data
    df = pd.read_csv(
        'https://raw.githubusercontent.com/mwaskom/seaborn-data/master/iris.csv')

    train(df)

    # Randomly tweak numerical data
    df_numerical = df.select_dtypes(include=[np.number])
    df[df_numerical.columns] = df_numerical.apply(
        lambda x: x*uniform(0.8, 1))

    predict(df)


if __name__ == '__main__':
    run()



import os
import sys

from minotor.data_managers.file_manager import FileManager
from minotor.run import runserver
from minotor.seeds.iris_seed import run as run_iris

fm = FileManager()

ARG_COMMAND = {
    'iris-example': run_iris,
    'clean': fm.clean_data,
    'clean-training': fm.clean_feature_data,
    'clean-prediction': fm.clean_prediction_data,
}


def main():
    # if no args, then runserver
    if len(sys.argv) <= 1:
        return runserver()

    # keeping only the first arg
    command_name = sys.argv[1]
    if command_name not in ARG_COMMAND:
        available_command = ' '.join(list(ARG_COMMAND.keys()))
        if len(ARG_COMMAND.keys()) == 1:
            return f"The only available command after minotor is: {available_command}"
        return f"Available commands after minotor are: {available_command}"
    return ARG_COMMAND.get(command_name)()


if __name__ == '__main__':
    main()



from dataclasses import dataclass, field
from typing import List, Dict

from minotor.data_managers.data_containers.feature_phase_container import FeaturePhaseContainerABC, \
    FeaturePredictionPhaseContainer, \
    FeatureTrainingPhaseContainer
from minotor.data_managers.data_types import DataType
from minotor.statistics.statistic_library import statistic_library, StatisticLibrary, crossed_statistic_library


@dataclass
class SingleFeatureDataContainer:
    data_type: DataType
    training_phase: FeaturePhaseContainerABC = field(default_factory=FeatureTrainingPhaseContainer)
    prediction_phase: FeaturePhaseContainerABC = field(default_factory=FeaturePredictionPhaseContainer)

    @staticmethod
    def from_json(data: Dict):
        return SingleFeatureDataContainer(data_type=DataType.type2value(data["type"]),
                                          training_phase=FeatureTrainingPhaseContainer.from_json(data["training"]),
                                          prediction_phase=FeaturePredictionPhaseContainer.from_json(data["prediction"])
                                          )

    def get_dict(self):
        return {
            "type": self.data_type.value,
            "training": self.training_phase.get_dict(),
            "prediction": self.prediction_phase.get_dict()
        }

    def update_feature(self, data: Dict, is_training: bool):
        phase_data = self.training_phase if is_training else self.prediction_phase
        phase_data.add_values(data)
        phase_data.compute_statistics(statistic_library)
        self.compute_crossed_statistics(crossed_statistic_library)

    def compute_crossed_statistics(self, crossed_statistic_library: StatisticLibrary):
        if self.training_phase.values and self.prediction_phase.values:
            for statistic_name, result in crossed_statistic_library.compute_all_statistics(
                    [self.training_phase.list_values, list(self.prediction_phase.values.values())]
            ):
                self.prediction_phase.statistics[statistic_name] = result



from minotor.data_managers.file_manager import FileManager
from minotor.api.base_route import BaseRouteHandler

class DataHandler(BaseRouteHandler):
  def get(self):
    fm = FileManager()
    cached_data = fm.get_json()
    self.set_header('Content-Type', 'application/json')
    self.write(cached_data)
    self.set_status(200)


from pathlib import Path

PACKAGE_PATH = Path(__file__).absolute().parent

DATA_DIR = Path.home() / ".cache" / "pyminotor"
DATA_DIR.mkdir(parents=True, exist_ok=True)

DATETIME_FORMAT = "%m/%d/%y %H:%M:%S"
DATETIME_ID_FORMAT = "%m%d%y%H%M%S%f"
